/*
 * TuxNES
 *
 * Written by Quor as part of the Nestra codebase
 * Extensively modified by The Mighty Mike Master
 * Also modified by Benjamin C. W. Sittler for TuxNES
 *
 * Description: This file handles the output of the graphics and the input
 * from the keyboard and the joystick.
 *
 * $Id: dc.c,v 1.10 2001/12/14 05:50:49 jim Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#undef USE_TA

#ifndef HAVE_DC
#ifdef HAVE_FEATURES_H
#include <features.h>
#endif /* HAVE_FEATRES_H */

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#else  /* !HAVE_DC */
#include "dc-glue.h"
#include <kos.h>
#endif /* !HAVE_DC */

#include "consts.h"
#include "globals.h"
#include "mapper.h"
#include "renderer.h"
#include "sound.h"

/* imports */
void	quit(void);
void	START(void);
void	fbinit(void);

#ifdef HAVE_DC
#define HAVE_DC_TIME
// Should be able to remove above line, redundant

// FIXME
#undef fprintf
#define fprintf(ARGS...) ;
/* exported functions */
int	InitDisplayDC(int argc, char **argv);
void	UpdateColorsDC(void);
int	UpdateDisplayDC(void);

static void	InitScreenshotDC(void);
static void	SaveScreenshotDC(void);
static int	HandleMapleDC(void);
extern int	screenshotnumber;
extern char	*screenshotfile;
extern unsigned char	needsredraw;		/* Refresh screen display */
extern unsigned char	redrawbackground;	/* Redraw tile background */
extern unsigned char	redrawall;		/* Redraw all scanlines */
extern unsigned int	currentbgcolor, oldbgcolor;

uint16 *bufferDC[2] = {0, 0};
int frameno;

uint16 whitepixel, blackpixel;

uint16 paletteDC[64], palette2DC[64];
uint16 basepixel;

/* TuxNES can only use 2 controllers */
#define MAX_CONTROLLERS 2
uint8 maple_controller[MAX_CONTROLLERS];

uint8 maple_nth_device(int code, int n);

#  ifdef USE_TA
uint32 txraddr;
void draw_screen_poly();
#  endif

#endif

#ifdef HAVE_DC

void
InitScreenshotDC(void)
{
}

void
SaveScreenshotDC(void)
{
}

/* First with a given function code... */
uint8 maple_nth_device(int code, int n) {
	int port, unit;
	int i = 0;

	for (port=0; port<4; port++) {
		for (unit=0; unit<6; unit++) {
			if (maple_device_func(port, unit) & code) {
				if (i == n)
					return maple_create_addr(port, unit);
				i++;
			}
		}
	}
	return 0;
}

int
HandleMapleDC(void)
{
	// FIXME this means all buttons are depressed
	static unsigned int last_cond[MAX_CONTROLLERS] = { 0 };
	int i;
	cont_cond_t cond;

	fprintf(stderr,"handlemapledc: enter\n");

	for (i = 0; i < MAX_CONTROLLERS; i++) {

		if (!maple_controller[i]) return 0;

		if (cont_get_cond(maple_controller[i], &cond)) {
			fprintf(stderr, "Error getting controller %d status\n", i);
			/* libdream cleanup? */
			return 1;
		}

		/* START+B to quit */
		if (!(cond.buttons & (CONT_START | CONT_B))) {
			fprintf(stderr,"escape sequence pressed, quitting\n");
			return 1;                      
		}

		/* Unisystem stuff not implemented */

		fprintf(stderr,"cond.buttons: %04x\n", cond.buttons);
		controller[i] = 0;
		if (!(cond.buttons & CONT_START)) { controller[i] |= STARTBUTTON; }
		if (!(cond.buttons & CONT_A)) { controller[i] |= BUTTONA; }
		if (!(cond.buttons & CONT_X)) { controller[i] |= BUTTONB; }
		if (!(cond.buttons & CONT_DPAD_LEFT)) { controller[i] |= LEFT; }
		if (!(cond.buttons & CONT_DPAD_RIGHT)) { controller[i] |= RIGHT; }
		if (!(cond.buttons & CONT_DPAD_UP)) { controller[i] |= UP; }
		if (!(cond.buttons & CONT_DPAD_DOWN)) { controller[i] |= DOWN; }
		if (!(cond.buttons & CONT_Y)) { controller[i] |= SELECTBUTTON; }
#if 0
		if (!(cond.buttons & CONT_B)) { 
			if ((last_cond[i] & CONT_B)) { total_temp_ctr = 1; }
		} else {
			if (!(last_cond[i] & CONT_B)) { total_temp_ctr = 0; }
		}
#endif
		last_cond[i] = cond.buttons;
	}
	fprintf(stderr,"handlemapledc: leave\n");
	return 0;

}

int
InitDisplayDC(int argc, char **argv)
{
  int x;

/*  struct timeval time; */

  if (magstep < 1) {
    magstep = 1;
  }

  if (magstep > maxsize) {
    fprintf (stderr, "Warning: Enlargement factor %d is too large!\n",
	     magstep);
    magstep = maxsize;
  }

	/* ensure hardcode magstep for now */
	magstep = 1;
#ifdef USE_TA
	vid_init(DM_640x480, PM_RGB565);
	vid_empty();
	ta_init();
#else
	if (magstep == 2) 
		vid_init(DM_640x480, PM_RGB565);
	else
		vid_init(DM_320x240, PM_RGB565);
	vid_empty();  /* vid_init only clears the first page */
#endif

      indexedcolor = 0;
      blackpixel = 0x0000;
      whitepixel = 0xffff;
      for (x=0; x < 24; x++)
	  palette[x] = palette2[x] = blackpixel;

      for (x=0; x < 64; x++)
	{
	  /* RGB565 */
	  uint8 r,g,b;
	  r = (NES_palette[x] >> 16 >> 3); 
	  g = (NES_palette[x] >>  8 >> 2);
	  b = (NES_palette[x] >>       3);
	  paletteDC[x] = (r << 11) | (g << 5) | (b);
	  palette2DC[x] = blackpixel;
	}

      if (scanlines && (scanlines != 100))
      {
	/* ensure scanlines < 100 ?? */
	/* Kinda redundant on a TV, but maybe not on a VGA box */
	for (x = 0; x < 64; x++)
	  {
	    uint8 r,g,b;
	    
	    r = (uint8)((((NES_palette[x] & 0xFF0000) >> 16) * (scanlines / 100.0))) >> 3;
	    g = (uint8)((((NES_palette[x] & 0xFF00)   >>  8) * (scanlines / 100.0))) >> 2;
	    b = (uint8)((((NES_palette[x] & 0xFF)          ) * (scanlines / 100.0))) >> 3;

	    palette2DC[x] = (r << 11) | (g << 5) | (b);
	  }
      }



  bpp = 16;
  if (magstep == 2) { /*magstep = 2;*/ }
  scanlines = 100;

#ifdef USE_TA
  bytes_per_line = 2 * 512;
	ta_bkg.flags1 = 0x90800000;
	ta_bkg.flags2 = 0x20800440;
	ta_bkg.dummy = 0;
	ta_bkg.x1 = 0.0f;
	ta_bkg.y1 = 480.0f;
	ta_bkg.z1 = 0.2f;
	ta_bkg.argb1 = 0xff000000;
	ta_bkg.x2 = 0.0f;
	ta_bkg.y2 = 0.0f;
	ta_bkg.z3 = 0.2f;
	ta_bkg.argb2 = 0xff000000;
	ta_bkg.x3 = 640.0f;
	ta_bkg.y3 = 480.0f;
	ta_bkg.z3 = 0.2f;
	ta_bkg.argb3 = 0xff000000;

	txraddr = 0x700000; // txr_alloc(512*256*2);
	bufferDC[0] = (uint16 *)ta_texture_map(txraddr) + 8;
	rfb = (char *)bufferDC[0];
	fb = rfb;

#else
  bytes_per_line = 2 * 320 * magstep;
  bufferDC[0] = vram_s;
  bufferDC[1] = vram_s + 320 * magstep * 240 * magstep;
  frameno = 0;
  vid_waitvbl(); 
  vid_set_start((uint32) bufferDC[frameno]); 
  frameno ^= 1; 
  rfb = (char *)(bufferDC[frameno]) + ((magstep == 1 ? 320 : 640) - 256); // temporary kludge to center (see below)
  fb  = rfb;
#endif

  bpu = 8;       /* ?? */
  pix_swab = 0;  /* 0 for little-endian */
  lsb_first = 0; /* only matters for 1 or 4bpp */
  lsn_first = 0; /* only matters for 4 bpp */
  depth = 16;    /* not used outside this file, but global */

  if (verbose)
    {
      fprintf (stderr, "DC: %dbpp, mode %s\n", bpp, magstep == 2 ? "PM_640x480" : "PM_320x240");
    }
	for (x = 0; x < MAX_CONTROLLERS; x++) {
		maple_controller[x] = maple_nth_device(MAPLE_FUNC_CONTROLLER,x);
		fprintf(stderr,"DC: maple_controller[%d] = %02x\n", x, maple_controller[x]);
	}
  InitScreenshotDC ();
  fbinit ();
  fprintf(stderr,"DC: fbinit() completed\n");

#ifndef HAVE_DC_TIME
  gettimeofday (&time, NULL);
  renderer_data.basetime = time.tv_sec;
#else
	timer_prime(2,1,0);  /* second arg is junk */
	timer_clear(2);

#define TIMER32(o) ( *((volatile uint32*)(0xffd80000 + (o))) )
#define TCOR2 0x20
#define TCNT2 0x24

    /* kos doesn't let us set the timer to (uint)-1 */
	TIMER32(TCOR2) = 0xffffffff; /* TCOR0 */
	TIMER32(TCNT2) = 0xffffffff; /* TCNT0 */

	timer_start(2);
    renderer_data.basetime = timer_count(2);
#endif
  fprintf(stderr,"DC: initdisplay finished\n");
  return 0;
}

/* Update the colors on the screen if the palette changed */
void
UpdateColorsDC(void)
{
  int x;

  fprintf(stderr,"DC: updatecolors begin\n");

  /* Set Background color */
  /* FIXME The commented out stuff should be removed */
//  oldbgcolor = currentbgcolor;
      palette[24] = currentbgcolor = paletteDC[VRAM[0x3f00] & 63];
//      if (scanlines && (scanlines != 100))
//	palette2[24] = palette2DC[VRAM[0x3f00] & 63];
//     if (oldbgcolor != currentbgcolor /*||currentbgcolor!=bgcolor[currentcache] */ )
//	{
//	  redrawbackground = 1;
//	  needsredraw = 1;
//	}
  /* Set palette tables */
      for (x = 0; x < 24; x++)
	{
	  palette[x] = paletteDC[VRAM[0x3f01 + x + (x / 3)] & 63];
//	  if (scanlines && (scanlines != 100))
//	    palette2[x] = palette2DC[VRAM[0x3f01 + x + (x / 3)] & 63];
	}
  fprintf(stderr,"DC: updatecolors end\n");
}

int
UpdateDisplayDC(void)
{
#ifndef HAVE_DC_TIME
  struct timeval time;
  unsigned int timeframe;
  unsigned static int frame;
#else
  unsigned static int frame = 0;
#endif
  int done;

  static int nodisplay = 0;

  /* do audio update */
  sounder->UpdateAudio();

  /* Check the time.  If we're getting behind, skip a frame to stay in sync. */
#ifndef HAVE_DC_TIME
  gettimeofday (&time, NULL);
  timeframe = (time.tv_sec - renderer_data.basetime) * 50 + time.tv_usec / 20000;     /* PAL */
  timeframe = (time.tv_sec - renderer_data.basetime) * 60 + time.tv_usec / 16666;     /* NTSC */
  frame++;
  if (halfspeed)
    timeframe >>= 1;
  if (doublespeed == 2)
    timeframe <<= 1;
  else if (doublespeed > 2)
    timeframe *= doublespeed;
  if (desync)
    frame = timeframe;
  desync = 0;
  if (frame < timeframe - 20 && frame % 20 == 0)
    desync = 1;                 /* If we're more than 20 frames behind, might as well stop counting. */
#else
  total_frames++;
  if (total_frames % (60*15) == 0) {
	  printf(".");
	}
#endif

/*  fprintf(stderr,"updatedisplay, returning immediately\n"); */
  fprintf(stderr,"updatedisplay begin, frame %d\n", total_frames);
/*  fprintf(stderr,"cksum_add32: %04x\n",cksum_add32(ROM,40976)); */
/*  if (cksum_add32(ROM,40976) != 0x40157a) {
        fprintf(stderr, "cksum error\n");
      	quit();
  } */
  if (! nodisplay)
    {
//	frameskip = 1;
      drawimage (PBL);
      /* Is frameskip placed correctly?  Page flipping is negligible
       * only the VBL wait is timeful. */
      if (!frameskip)
	{
	  UpdateColorsDC();

      vid_waitvbl();
#ifdef USE_TA
{
	poly_hdr_t poly;
	ta_begin_render();
    ta_build_poly_hdr(&poly, TA_OPAQUE, TA_RGB565, 512, 256, txraddr, TA_NO_FILTER);
    ta_commit_poly_hdr(&poly);

	draw_screen_poly();

	ta_commit_eol();

    ta_build_poly_hdr(&poly, TA_TRANSLUCENT, TA_NO_TEXTURE, 0, 0, 0, 0);
	ta_commit_poly_hdr(&poly);

	ta_commit_eol();
	ta_finish_frame();

	rfb = (char *)bufferDC[0];
	// Since we don't use two buffers, we might flicker
	fb = rfb;
}
	
#else
{
	int i;
	uint32 *ptr1 = (uint32 *)bufferDC[frameno] + 12;
	for (i = 0 ; i < 240; i++) {
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  ptr1 += 128;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  *ptr1++ = 0;
	  ptr1 += 24;
	}
}

	  vid_set_start((uint32) bufferDC[frameno]); 
	  frameno ^= 1;
	  rfb = (char *)(bufferDC[frameno]) + ((magstep == 1 ? 320 : 640) - 256); //temporary kludge to center
          fb =  rfb;
//  sounder->UpdateAudio();

#endif
	  redrawall = needsredraw = 0;
	} else { /* !frameskip */
//      vid_waitvbl();
//  sounder->UpdateAudio();
    }
	}
  else {
      fprintf(stderr,"nodisplay this frame\n");
  }

  frameskip = frameskip + 1;
  if (frameskip > 1) frameskip = 0; /**/

  needsredraw = 0;
  redrawbackground = 0;
  redrawall = 0;

#ifndef HAVE_DC_TIME

  /* Slow down if we're getting ahead */
  if (frame > timeframe + 1 && frameskip == 0)
    {
      usleep (16666 * (frame - timeframe - 1));
    }
#endif

//if (total_frames > 4000) { quit(); }

  /* Input loop */
//  do {
    done = HandleMapleDC();
//  } while (renderer_data.pause_display);  /* Should probably remove this test */
  
#ifndef HAVE_DC_TIME
  /* Check the time.  If we're getting behind, skip next frame to stay in sync. */
  gettimeofday (&time, NULL);
  timeframe = (time.tv_sec - renderer_data.basetime) * 60 + time.tv_usec / 16666;     /* NTSC */
  if (halfspeed)
    timeframe >>= 1;
  if (doublespeed == 2)
    timeframe <<= 1;
  else if (doublespeed > 2)
    timeframe *= doublespeed;
  if (frame >= timeframe || frame % 20 == 0)
    frameskip = 0;
  else
    frameskip = 1;
#endif
  fprintf(stderr,"end updatedisplay\n");
  return done;
}

#ifdef USE_TA
void draw_screen_poly()
{
	vertex_ot_t vert;
	const int x = (640 - (256 * 2)) / 2;   // 64

	vert.flags = TA_VERTEX_NORMAL;
	vert.x = x + 0.0f;
	vert.y = 480.0f;
	vert.z = 1.0f;
	vert.u = 8.0f/512.0f; vert.v = 480.0f/512.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_NORMAL;
	vert.y = 0.0f;
    vert.v = 0.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_NORMAL;
	vert.x = x + 512.0f;
	vert.y = 480.0f;
	vert.u = 264.0f/512.0f; vert.v = 480.0f/512.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));

	vert.flags = TA_VERTEX_EOL;
	vert.y = 0.0f;
	vert.v = 0.0f;
	vert.a = 1.0f; vert.r = 1.0f; vert.g = 1.0f; vert.b = 1.0f;
	ta_commit_vertex(&vert, sizeof(vert));
}

#endif

#endif
